/*
 * Copyright (c) Sybase, Inc. 2002   
 * All rights reserved.                                    
 * 
 * This file contains Original Code and/or Modifications of 
 * Original Code as defined in and that are subject to the  
 * Sybase Public Source License version 1.0 (the           
 * "License").  You may not use this file except in         
 * compliance with the License.  Please obtain a copy of    
 * the License at http://www.sybase.com/developer/opensource 
 * and read it before using this file.                               
 * 
 * The Original Code and all software distributed under the 
 * License are distributed on an 'AS IS' BASIS, WITHOUT     
 * WARRANTY OF ANY KIND, EITHER EXPRESS OR IMPLIED, AND     
 * SYBASE HEREBY DISCLAIMS ALL SUCH WARRANTIES, INCLUDING  
 * WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,    
 * FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR     
 * NON-INFRINGEMENT.                                        
 * 
 * Please see the License for the specific language.         
 */
package org.kldp.ebmsh.message.packaging;

import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.Random;

import javax.mail.MessagingException;
import javax.mail.internet.ContentType;
import javax.mail.internet.InternetHeaders;
import javax.mail.internet.MimeBodyPart;
import javax.mail.internet.MimeMultipart;
import javax.xml.soap.SOAPException;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.kldp.ebmsh.message.util.ByteArrayDataSource;
import org.kldp.ebmsh.message.util.SOAPConstants;



/**
 * ebXML context for a message. Encapsulates:
 * <ul>
 *  <li>MIME multipart
 * </ul>
 *
 * @author Cliff Collins
 */
public class EBXMLMimeMultipart
{
    
	/**
	@since 2003.1.1
	@roseuid 400B45890162
	 */
	private static Log _log = LogFactory.getLog(EBXMLMimeMultipart.class); 


    private static final String HEXMAP = "0123456789ABCDEF";	// map for converting chars into hex rep.

    private MimeMultipart		_mimeParts;
    private ContentType			_ct;

    public EBXMLMimeMultipart(ByteArrayDataSource bas)
        throws SOAPException
    {
        boolean 	bSinglePart = false;
        boolean 	bDecrypted = false;

        // Look at the bas ct...if not multipart but just text/xml, then it 
        // could be a soap fault. turn this into the first body part of a multipart message
        try
        {
            ContentType	ct = new ContentType(bas.getContentType());

            // If the root type is "text/xml" it might be a soap fault (this may not be 
            // a multipart). We should build it into a multipart and let our normal processing code 
            // handle it. If it isn't a soap fault, the normal parsing code will error on the message.
            // since it won't have a SOAP envelope.
            bSinglePart = SOAPConstants.HEADERVAL_CONTENT_TYPE.equalsIgnoreCase(ct.getBaseType());
            bDecrypted = SOAPConstants.HEADERVAL_CONTENT_TYPE_MULTIPART.equalsIgnoreCase(ct.getBaseType());
        }
        catch(Exception ee)
        {
        }

        if (bSinglePart) 
        {
            buildMultipartFromSinglePart(bas);
        }
        else
        {
            buildMultipart(bas, bDecrypted);
        }
        checkMimeMultipart();
    }

    /**
	 * Method to create the MimeBodyPart corresponding tho ebXML H Container
	 * @param EBXMLEnvelopeBody  is the string corresponding to the EBXMLEnvelopeBody
	 */
    private void buildMultipartFromSinglePart(ByteArrayDataSource bas)
        throws SOAPException
    {
        try
        {
			Random r = new Random();
            // Build a Content type for what it would be if it was a multipart. This will
            // allow it to pass all the checks on the multipart object.
            _ct = new ContentType(SOAPConstants.HEADERVAL_CONTENT_TYPE_MULTIPART);
            _ct.setParameter(SOAPConstants.HEADER_TYPE, SOAPConstants.HEADERVAL_CONTENT_TYPE);
            _ct.setParameter(SOAPConstants.HEADER_START, SOAPConstants.MessageContentId);
            _ct.setParameter(SOAPConstants.HEADER_BOUNDARY, SOAPConstants.FAULT_SOAP_ACTION + String.valueOf(r.nextInt())+ SOAPConstants.HEADER_BOUNDARY);
            //_ct.setParameter(SOAPConstants.HEADER_BOUNDARY, "melong");

            InternetHeaders header = new InternetHeaders();
            header.addHeader(SOAPConstants.HEADER_CONTENT_TYPE, bas.getContentType());
            header.addHeader(SOAPConstants.HEADER_CONTENT_ID, SOAPConstants.MessageContentId);

            MimeBodyPart	mbp = new MimeBodyPart(header, bas.toByteArrayTrim());

            _mimeParts = new MimeMultipart(SOAPConstants.HEADERVAL_MULTIPART_CONTENT_SUBTYPE);
            _mimeParts.addBodyPart(mbp, 0);	
        }
        catch (IOException ee)
        {
            throw new SOAPException("< " + SOAPConstants.FAULT_CODE_PROTOCOL +  " >Badly formated MIME message.");
        }
        catch (javax.mail.MessagingException ee)
        {
            throw new SOAPException("<"+  SOAPConstants.FAULT_CODE_PROTOCOL +">" + "Badly formated MIME message.");
        }
    }

    private void buildMultipart(ByteArrayDataSource bas, boolean bDecrypted)
        throws SOAPException
    {
       

        try
        {
            if (bDecrypted)
            {
                // If the message is S/MIME then we need to call the security MimeMultipart.
                _mimeParts = new MimeMultipart(bas);
            }
         

            _ct = new ContentType(_mimeParts.getContentType());
        }
       
        catch (javax.mail.MessagingException ee)
        {
            throw new SOAPException( "Badly formated MIME message.", ee);
        }
       
        
    }

    private void checkMimeMultipart()
        throws SOAPException
    {
        try
        {
            // check the content type header for all the required fields.
            String	ct_type = _ct.getParameter(SOAPConstants.HEADER_TYPE);
            String	ct_baseType = _ct.getBaseType();
            String	ct_boundary = _ct.getParameter(SOAPConstants.HEADER_BOUNDARY);
            String	ct_startID = _ct.getParameter(SOAPConstants.HEADER_START);

            // Check the fields.
            if (SOAPConstants.HEADERVAL_CONTENT_TYPE_MULTIPART.equalsIgnoreCase(ct_baseType) == false)
            {
                throw new SOAPException( "Invalid base type=" + ct_baseType);
            }
            if (SOAPConstants.HEADERVAL_CONTENT_TYPE.equalsIgnoreCase(ct_type) == false)
            {
                throw new SOAPException( "Invalid content type=" + ct_type);
            }
            if ( (ct_boundary == null) || (ct_boundary.length() == 0))
            {
                throw new SOAPException( "Empty MIME Boundary.");
            }

            // If there is a payload than the first item should have a 
            // Content-ID equal to the "start=" field in the mime header!
            if ( (ct_startID != null) && (ct_startID.length() > 0) && (getCount() > 0) )
            {
                MimeBodyPart	firstBP = getBodyPart(0);	// get the first payload and make sure it has the correct CID
                String			firstCID = firstBP.getContentID();
                if (firstCID != null)
                {
                	String bracketStartID = ct_startID;
    				if ( (ct_startID.startsWith("<") == false) && (ct_startID.endsWith(">") == false) )
    				{
    					bracketStartID = '<' + ct_startID + '>';
    				}
 
                	// try and be really flexable about what they might put in the cid text.
                	if ( (ct_startID.equals(firstCID) == false) && 			// basic compare...assume no encoding and no <>
                		 (bracketStartID.equals(firstCID) == false) &&		// <cid>
                		 (ct_startID.equals(decodeURL(firstCID)) == false) &&	// decode the cid and compare
                		 (bracketStartID.equals(decodeURL(firstCID)) == false) )	// decoded with <>
                	{	
                		_log.info( "Content 'start' field(" + ct_startID + ") not equal to first body part's Content-ID(" + firstCID + ").");
                    	// According to the spec this needs to be true for robust error detection. It is not a required field
                    	// so if the field is not set it is not an error but if the field is set it must match!
                    	throw new SOAPException( "Content 'start' field not equal to first body part's Content-ID.");
                	}
                }
            }
        }
        catch (SOAPException ee)
        {
            throw ee;
        }
        catch (javax.mail.MessagingException ee)
        {
            throw new SOAPException("Failed to parse MIME data.", ee);
        }
        catch (Throwable ee)
        {
            // trp.ebxml.log.0009=Parsing new message got an unexpected error.
            _log.error("trp.ebxml.log.0009", ee);
            // return a soap fault error to the sender.
            throw new SOAPException(SOAPConstants.FAULT_CODE_SERVER +
                ".Exception:", ee);
        }
    }

    /**
     * Get the specified Part.  Parts are numbered starting at 0.
     *
     * @param index     the index of the desired Part
     * @return          the Part, or null if no such part exists.
     */
    public MimeBodyPart getBodyPart(int index)
    {
        /*
		* NOTE: getBodyPart method never throws a MessagingException. In case a
		* future implementation does, catch it and throw an
		* IndexOutOfBoundsException
		*/
        if (_mimeParts == null)
        {
            return null;
        }

        try
        {
            return (MimeBodyPart)_mimeParts.getBodyPart(index);
        }
        catch (MessagingException me)
        {
            throw new IndexOutOfBoundsException(me.getMessage());
        }
        catch (IndexOutOfBoundsException iobe)
        {
            return null;
        }
        catch (NullPointerException npe)
        {
            return null;
        }
    }

    /**
     * Get the Mimepart referred to by the given ContentID (CID).
     * Returns null if the part is not found.
     *
     * @param  CID      the ContentID of the desired part
     * @return          the Part, or null if no such part exists.
     */
    public MimeBodyPart getBodyPart(String CID)
    {
        if (_mimeParts == null)
        {
            return null;
        }

        try
        {
            return (MimeBodyPart)_mimeParts.getBodyPart(CID);
        }
        catch (MessagingException me)
        {
            return null;
        }
        catch (NullPointerException npe)
        {
            return null;
        }
        catch (IndexOutOfBoundsException iobe)
        {
            return null;
        }
    }

    /**
     * Find the Mimepart referred to by the given URI. This can be:<ul>
     * <li>An absolute URI, in which case a part is located with a
     *     corresponding Content-Location header.
     * <li>A relative URI, in which case an absolute URI is constructed
     *     relative to the Content-Location header of the root SOAP part
     *     or the multipart and the previous rule is then applied.
     * <li>A URI of the format "cid:xxx"> in which case a part is located
     *     with a matching Content-ID header.
     * </ul>Returns null if the part is not found.
     * <p>Note: relative URIs not entirely implemented yet.
     *
     * @param  uri      the URI
     * @return          the Part or null if not found
     */
    public MimeBodyPart findBodyPart(String uri)
    {
        if (_mimeParts == null || uri == null)
        {
            return null;
        }

        try
        {
            if ( (uri.length() > 4) && uri.substring(0, 4).equalsIgnoreCase(SOAPConstants.CID_FIELD) )
            {
                // It's a Content-ID. URLDecode and lookup.
                String cid = decode(uri.substring(4));

                try
                {
                	// do we need to try "<encoded(cid)>" ?????
                    return (MimeBodyPart)_mimeParts.getBodyPart(cid);
                }
                catch(NullPointerException npe)
                {
                }
            }
            else
            {
                // It's a URI.
                return findPartByLocation(uri);
            }

        }
        catch (MessagingException me)
        {
        }
        catch (NullPointerException npe)
        {
        }
        return null;
    }

    /**
     * Find the Mimepart referred to by the given URI. This can be:<ul>
     * <li>An absolute URI, in which case a part is located with a
     *     corresponding Content-Location header.
     * <li>A relative URI, in which case an absolute URI is constructed
     *     relative to the Content-Location header of the root SOAP part
     *     or the multipart and the previous rule is then applied.
     * </ul>Returns null if the part is not found.
     * <p>Note: relative URIs not entirely implemented yet. We also can't pick
     * up the base URI from the multipart (transport) headers yet, only from
     * the root (SOAP) part.
     *
     * @param  uri      the URI
     * @return          the Part or null if not found
     */
    public MimeBodyPart findPartByLocation(String uri)
    {
        try
        {
            String baseUri = getBaseURI();
            uri = normalizeURI(uri, baseUri);
            if (uri == null)
            {
                return null;
            }

            for (int ii = 0; ii < _mimeParts.getCount(); ii++) 
            {
                MimeBodyPart part = getBodyPart(ii);

                if (part != null)
                {
                	// do we need to try "<encoded(cid)>" ?????
 
                    // SOAP w/ attachments says either header type can be used for location.
                    String partUri = part.getContentID();
                    if (partUri == null)
                    {
                        partUri = part.getHeader(SOAPConstants.HEADER_CONTENT_LOCATION, null);
                    }
                    if (uri.equals(normalizeURI(partUri, baseUri)))
                    {
                        return part;
                    }
                }
            }
        }
        catch(MessagingException me)
        {
        }
        return null;
    }

    /**
     * Determine the document's base URI, according to the following scheme:
     * <ol>
     * <li>The Content-Location header of the root (SOAP) part.
     * <li>The Content-Location header of the multipart. (not implemented)
     * <li>"thismessage:/"
     * </ol>
     */
    public String getBaseURI()
    {
        String baseUri = null;

        try
        {
            // SOAP w/ attachments says either header type can be used for location.
            baseUri = getRootPart().getContentID();
            if (baseUri == null)
            {
                baseUri = getRootPart().getHeader(SOAPConstants.HEADER_CONTENT_LOCATION, null);
            }
        }
        catch(MessagingException me)
        {
        }
        if (baseUri == null)
        {
            baseUri = SOAPConstants.DEFAULT_BASEURI;
        }
        return baseUri;
    }

    /**
     * Find the root part. For multipart, search for a Constants.HEADER_START Content-Type
     * header modifier, if not present or invalid, assume the first part is
     * the root.
     *
     * @return          document root BodyPart
     * @exception       MessagingException
     */
    public MimeBodyPart getRootPart()
        throws MessagingException
    {
        MimeBodyPart rootPart = null;

        if (getCount() > 1)
        {
            String startCid = _ct.getParameter(SOAPConstants.HEADER_START);
            
            if (startCid != null)
            {
                rootPart = getBodyPart(decode(startCid));
            }
        }
        if (rootPart == null)
        {
            rootPart = getBodyPart(0);
        }
        return rootPart;
    }

    /**
	 * Return the content type set from the HTTP header.
	 */
    public ContentType getContentType()
    {
        if (_ct == null)
        return null;

        return _ct;
    }

    /**
	 * Return the content type set from the HTTP header.
	 */
    public String getContentTypeString()
    {
        if (_mimeParts != null)
        return null;

        return _mimeParts.getContentType();
    }

    /**
	 * Return the number of enclosed BodyPart objects.
	 *
	 * @return          number of parts
	 */
    public int getCount()
        throws MessagingException
    {
        if (_mimeParts == null)
        {
            return 0;
        }
        else
        {
            return _mimeParts.getCount();
        }
    }

    /**
     * URLDecode string.
     */
    public static String decode(String ss)
    {
        StringBuffer	ret = new StringBuffer();
        char			cc;
        for (int ii = 0; ii < ss.length(); ii++)
        {
            cc = ss.charAt(ii);
            switch(cc)
            {
                case '+':
                    ret.append(' ');
                    break;
                case '%':
                    ret.append((char)((HEXMAP.indexOf(ss.charAt(++ii)) << 4)
                        + HEXMAP.indexOf(ss.charAt(++ii))));
                    break;
                default:
                    ret.append(cc);
            }
        }
        return ret.toString();
    }

    /**
     * Normalise URI to an absolute URI according to rules in RFC2396.
     * <p>Note: resolution for relative URIs is not completely implemented
     * yet. For now, we just concatenate the relative URI to the base URI.
     */
    public static String normalizeURI(String uri, String baseUri)
    {
        // Check if it is an absolute URI.
        int p1 = uri.indexOf(':');
        if (p1 >= 0)
        {
            String scheme = uri.substring(0, p1);
            if (scheme.indexOf('/') == -1 &&
            scheme.indexOf('?') == -1 &&
                scheme.indexOf('#') == -1)
                // Yes, it's absolute! Return unchanged.
                return uri;
        }
        // It's a relative URI. Normalise to an absolute URI.
        return baseUri + uri;
    }

    /**
    * String representation for debug purposes.
    */
    public String toString()
    {
        StringWriter sw = new StringWriter();
        PrintWriter pw = new PrintWriter(sw);

        pw.print("[Parts={");

        if (_mimeParts != null)
        {
            try
            {
                for (int ii = 0; ii < getCount(); ii++)
                {
                    if (ii > 0)
                    {
                        pw.print(", ");
                    }

                    MimeBodyPart mbp = getBodyPart(ii);
                    pw.print("[cid:" + mbp.getContentID()
                        + " type: " + mbp.getContentType()
                        + " enc: " + mbp.getEncoding() + "]");
                }
            }
            catch(MessagingException me)
            {
                // trp.ebxml.log.0001={0}
                _log.info(me.getMessage());
            }
        }

        pw.print("}]");

        return sw.toString();
    }
    
 
	/** 
		* Decode the URL string
		*/
	   public static String decodeURL(String theURL)
	   {
		   if (theURL == null)
		   {
			   return null;
		   }
    	
		   int				theSize = theURL.length();
		   int				kk;
		   StringBuffer	decoded = new StringBuffer(theSize);
		   char			ch;

		   for(int ii = 0; ii < theSize; ii++)
		   {
			   ch = theURL.charAt(ii);
			   if(ch == '%')
			   {
				   ch = theURL.charAt(++ii);
				   kk = Character.digit(ch, 16);
				   kk *= 16;
				   ch = theURL.charAt(++ii);
				   kk += Character.digit(ch, 16);
				   decoded.append((char)kk);
			   }
			   else if(ch == '+')
			   {
				   decoded.append(' ');
			   }
			   else
			   {
				   decoded.append(ch);
			   }
		   }

		   return decoded.toString();
	   }
	   
	   public MimeMultipart getMultiPart()
	   
	   {
	   	
	   	
	   	return this._mimeParts;
	   }
	   
   
}

