
package org.kldp.ebmsh.message.packaging.validation;

import java.util.ArrayList;
import java.util.Iterator;

import javax.xml.soap.AttachmentPart;
import javax.xml.soap.SOAPException;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.kldp.ebmsh.message.packaging.EbxmlBody;
import org.kldp.ebmsh.message.packaging.EbxmlMessage;
import org.kldp.ebmsh.message.packaging.ErrorCode;
import org.kldp.ebmsh.message.packaging.Reference;
import org.kldp.ebmsh.message.util.MessageUtil;
import org.kldp.util.log.LogUtil;

import com.sun.org.apache.xml.internal.utils.URI;
import com.sun.org.apache.xml.internal.utils.URI.MalformedURIException;

/**
 * File name   - BodyValidator.java
 * Create date - May 9, 2006
 * Update time - May 9, 2006
 * Version     - v1.0
 * 
 * Author      - Sirius
 */
public class BodyValidator {
	private static Log log = LogFactory.getLog(BodyValidator.class);
	
	private static BodyValidator instance = new BodyValidator();
	
	/**
	 * Constructor
	 * 
	 * BodyValidator class uses singlton pattern so constructor type is private.
	 */
	private BodyValidator() {
	}
	
	public static BodyValidator getInstance() {
		return instance;
	}

	/**
	 * Version     - v1.0
	 * Create time - May 9, 2007
	 * Update time - May 9, 2007
	 *
	 * Author      - Sirius
	 *
	 * @param ebxmlMessage
	 * @throws EbmsEbxmlValidationException
	 */
    public void validate(EbxmlMessage ebxmlMessage) throws EbmsEbxmlValidationException {
    	payloadValidation(ebxmlMessage);
    }
    
    
    /**
	 * Version     - v1.0
	 * Create time - May 7, 2007
	 * Update time - May 7, 2007
	 *
	 * Author      - Sirius
	 *
	 * @param ebxmlMessage
	 * @throws EbmsEbxmlValidationException
	 * 
	 * Error condition
	 * 1. Reference's xlink:href attribute contains a URI the is a content id then a MIME part with that content-id MUST be present in the corresponding Payload Container of the message.
	 * 2. Reference's xlink:href attribute contains a URI, not a content id, and the URI cannot be resolved.
	 * - It is not implements. bizbox not support external resource. 
	 */
	private void payloadValidation(EbxmlMessage ebxmlMessage) throws EbmsEbxmlValidationException {
		try {
			log.info(" Payload validation.");
			
			EbxmlBody ebxmlBody = ebxmlMessage.getEbxmlBody();
			
			if( !ebxmlBody.hasManifest() )
				return ;
			
			ArrayList references = ebxmlBody.getManifest().getReferences();
			ArrayList referenceContentIds = new ArrayList();
			
			for(int i = 0; i < references.size(); i++) {
				Reference reference = (Reference)references.get(i);
				
				String xlinkRef = reference.getXlinkRef();
				
				log.info("  Reference's XlinkRef check. XlinkRef : " + xlinkRef);
				
				if( !xlinkRef.startsWith("cid:") ) {
					log.fatal("   [ Error ] - XlinkRef not start 'cid:'.");
					
					throw new EbmsEbxmlValidationException(
							ErrorCode.MimeProblem,
							ErrorCode.Error,
							"No such cid in the reference element",
							"Payload validation");
				}
				
				if( !isValidUri(xlinkRef) ) {
					log.fatal("   [ Error ] - XlinkRef is not URI.");
					
					throw new EbmsEbxmlValidationException(
							ErrorCode.MimeProblem,
							ErrorCode.Error,
							"The xlink:href element of a Manifest/Reference element must be a URI.",
							"Payload validation");
				}
				
				referenceContentIds.add( xlinkRef.substring(4) );
			}
			
			
			Iterator iter = ebxmlMessage.getAttachments();
			ArrayList attachmentPartContentIds = new ArrayList();
			
			while( iter.hasNext() ) {
				AttachmentPart attachmentPart = (AttachmentPart)iter.next();
				
				String cid = attachmentPart.getContentId();

				if( cid.startsWith("<") )
					cid = cid.substring( cid.indexOf("<") + 1, cid.length() - 1 );
				
				attachmentPartContentIds.add(cid);
			}
			
			log.info("  Check reference's XlinkRef value existance in attachment part.");
			
			for(int i = 0; i < referenceContentIds.size(); i++) {
				String contentId = (String)referenceContentIds.get(i);
				
				if( !attachmentPartContentIds.contains(contentId) ) {
					log.fatal("   [ Error ] - A manifest entry refers to non-existent payload.");
					
					throw new EbmsEbxmlValidationException(
							ErrorCode.MimeProblem,
							ErrorCode.Error,
							"A manifest entry refers to non-existent payload.",
							"Payload validation");
				}
			}
		}
		catch(SOAPException e) {
			throw new EbmsEbxmlValidationException(ErrorCode.OtherXml, ErrorCode.Error, "Can't payload validation.\n" + LogUtil.stackTraceToString(e), "Payload validation");
		}
	}
	
	
	/**
	 * Version     - v1.0
	 * Create time - May 8, 2007
	 * Update time - May 8, 2007
	 *
	 * Author      - Sirius
	 *
	 * @param uri
	 * @return
	 * 
	 * Validate the uri.
	 */
	private boolean isValidUri(String uri) {
		try {
			new URI(uri);
			
			return true;
		}
		catch (MalformedURIException e) {
			return false;
		}
	}
}
