package org.kldp.ebmsh.message.packaging.validation;

import java.net.MalformedURLException;
import java.net.URL;
import java.util.Date;
import java.util.Iterator;

import javax.xml.soap.SOAPException;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.kldp.ebmsh.config.EbmsConstants;
import org.kldp.ebmsh.cpa.CanReceive;
import org.kldp.ebmsh.cpa.Cpa;
import org.kldp.ebmsh.cpa.CpaManagerFactory;
import org.kldp.ebmsh.cpa.DeliveryChannel;
import org.kldp.ebmsh.cpa.PartyInfo;
import org.kldp.ebmsh.cpa.ejb.CpaManager;
import org.kldp.ebmsh.cpa.ejb.CpaManagerHome;
import org.kldp.ebmsh.message.packaging.EbxmlConstants;
import org.kldp.ebmsh.message.packaging.EbxmlHeader;
import org.kldp.ebmsh.message.packaging.EbxmlMessage;
import org.kldp.ebmsh.message.packaging.ErrorCode;
import org.kldp.ebmsh.message.packaging.From;
import org.kldp.ebmsh.message.packaging.MessageOrder;
import org.kldp.ebmsh.message.packaging.PartyId;
import org.kldp.ebmsh.message.packaging.SequenceNumber;
import org.kldp.ebmsh.message.packaging.Service;
import org.kldp.ebmsh.message.util.MessageUtil;
import org.kldp.util.log.LogUtil;

/**
 * File name   - HeaderValidator.java
 * Create date - May 7, 2006
 * Update time - May 7, 2006
 * Version     - v1.0
 * 
 * Author      - Sirius
 */
public class HeaderValidator {
	private static Log log = LogFactory.getLog(HeaderValidator.class);
	
	private static HeaderValidator instance = new HeaderValidator();
	
	/**
	 * Constructor
	 * 
	 * HeaderValidator class uses singlton pattern so constructor type is private.
	 */
	private HeaderValidator() {
	}
	
	public static HeaderValidator getInstance() {
		return instance;
	}
	
	
	/**
	 * Version     - v1.0
	 * Create time - May 7, 2007
	 * Update time - May 7, 2007
	 *
	 * Author      - Sirius
	 *
	 * @param ebxmlMessage
	 * @throws EbmsEbxmlValidationException
	 */
	public void validate(EbxmlMessage ebxmlMessage) throws EbmsEbxmlValidationException {
		log.info("Start MessageHeader validation.");
		
		//timeToLiveValidation(ebxmlMessage);
		
		cpaValidation(ebxmlMessage);
		
		fromValidation(ebxmlMessage);
		
		toValidation(ebxmlMessage);
		
		actionValidation(ebxmlMessage);
		
		syncValidation(ebxmlMessage);
		
		serviceValidation(ebxmlMessage);
		
		ackRequestedValidation(ebxmlMessage);
		
		messageOrderValidation(ebxmlMessage);
		
		duplicateEliminationValidation(ebxmlMessage);
	}
	
	
	/**
	 * Version     - v1.0
	 * Create time - May 10, 2007
	 * Update time - May 10, 2007
	 *
	 * Author      - Sirius
	 *
	 * @param ebxmlMessage
	 * @throws EbmsEbxmlValidationException
	 * 
	 * Error condition
	 * 1. Can't find cpa.
	 */
	private void cpaValidation(EbxmlMessage ebxmlMessage) throws EbmsEbxmlValidationException {
		log.info(" CPA validation.");
			
		String cpaId = ebxmlMessage.getCPAId();
		
		Cpa cpa = new CpaManagerFactory().createInstance().getCpa(cpaId);

		if( cpa == null ) {
			log.fatal("    [ Error ] - Cpa: " + cpaId + " can't find from ebXML engine.");
			
			EbmsEbxmlValidationException ev = new EbmsEbxmlValidationException(
					ErrorCode.ValueNotRecognized,	
					ErrorCode.Error,
					"Check your cpaid, there's no such kinds of cpa id." + cpaId,
					"CPA Validation");

			throw ev;
		}
	}
	
	
	/**
	 * Version     - v1.0
	 * Create time - May 7, 2007
	 * Update time - May 7, 2007
	 *
	 * Author      - Sirius
	 *
	 * @param ebxmlMessage
	 * @throws EbmsEbxmlValidationException
	 * 
	 * Error condition
	 * 1. If 'PartyId''s 'type' attribute is not present, 'PartyId' element value MUST be a URI.
	 * 2. If can't find 'PartyInfo' from cpa coresponds to the ebXML message's From.  
	 */
	private void fromValidation(EbxmlMessage ebxmlMessage) throws EbmsEbxmlValidationException {
		try {
			log.info(" From element validation.");
			
			From from = ebxmlMessage.getEbxmlHeader().getMessageHeader().getFrom();
			
			Iterator iter = from.getPartyIds();
			
			if( !iter.hasNext() )
				throw new EbmsEbxmlValidationException(ErrorCode.inconsistent, ErrorCode.Error, "From PartyId is null.", "From validation");
			
			while( iter.hasNext() ) {
				PartyId partyId = (PartyId)iter.next();
				
				String type = partyId.getType();
				String value = partyId.getValue();
				
				log.info("  PartyId type: " + type + "\tvalue: " + value);
				
				if( type == null ) {
					log.info("   PartyId type is null. Check PartyId element value is URI.");
					
					if( !isValidUrl(value) ) {
						log.fatal("    [ Error ] - PartyId element value is not URI.");
						
						throw new EbmsEbxmlValidationException(ErrorCode.ValueNotRecognized, ErrorCode.Error, "From PartyId is not uri.", "From validation");
					}
				}
			}
			
			log.info("  Search the ParyInfo from CPA corresponds to the From of ebXML Message.");
			
			String cpaId = ebxmlMessage.getCPAId();
			String fromStr = ebxmlMessage.getFrom();
			
			log.info("   ebXML message From: " + fromStr);
			
			Cpa cpa = new CpaManagerFactory().createInstance().getCpa(cpaId);
			
			PartyInfo partyInfo = cpa.getPartyInfo(fromStr);
			
			if( partyInfo == null ) 
				throw new EbmsEbxmlValidationException(ErrorCode.ValueNotRecognized, ErrorCode.Error, "Can't find PartyInfo correspond to the From of ebXML Message.", "From validation");
		}
		catch(SOAPException e) {
			throw new EbmsEbxmlValidationException(ErrorCode.OtherXml, ErrorCode.Error, "Can't From element validation.\n" + LogUtil.stackTraceToString(e), "From validation");
		}
	}
	
	
	/**
	 * Version     - v1.0
	 * Create time - May 7, 2007
	 * Update time - May 7, 2007
	 *
	 * Author      - Sirius
	 *
	 * @param ebxmlMessage
	 * @throws EbmsEbxmlValidationException
	 * 
	 * Error condition
	 * 1. If 'PartyId''s 'type' attribute is not present, 'PartyId' element value MUST be a URI.
	 */
	private void toValidation(EbxmlMessage ebxmlMessage) throws EbmsEbxmlValidationException {
		try {
			log.info(" To element validation.");
			
			From from = ebxmlMessage.getEbxmlHeader().getMessageHeader().getFrom();
			
			Iterator iter = from.getPartyIds();
			
			if( !iter.hasNext() )
				throw new EbmsEbxmlValidationException(ErrorCode.inconsistent, ErrorCode.Error, "To PartyId is null.", "To validation");
			
			while( iter.hasNext() ) {
				PartyId partyId = (PartyId)iter.next();
				
				String type = partyId.getType();
				String value = partyId.getValue();
				
				log.info("  PartyId type: " + type + "\tvalue: " + value);
				
				if( type == null ) {
					log.info("   PartyId type is null. Check PartyId element value is URI.");
					
					if( !isValidUrl(value) ) {
						log.fatal("    [ Error ] - PartyId element value is not URI.");
						
						throw new EbmsEbxmlValidationException(ErrorCode.ValueNotRecognized, ErrorCode.Error, "To PartyId is not uri.", "To validation");
					}
				}
			}
			
			log.info("  Search the ParyInfo from CPA corresponds to the To of ebXML Message.");
			
			String cpaId = ebxmlMessage.getCPAId();
			String toStr = ebxmlMessage.getTo();
			
			log.info("   ebXML message To: " + toStr);
			
			Cpa cpa = new CpaManagerFactory().createInstance().getCpa(cpaId);
			
			PartyInfo partyInfo = cpa.getPartyInfo(toStr);
			
			if( partyInfo == null ) 
				throw new EbmsEbxmlValidationException(ErrorCode.ValueNotRecognized, ErrorCode.Error, "Can't find PartyInfo correspond to the To of ebXML Message.", "To validation");
		}
		catch(SOAPException e) {
			throw new EbmsEbxmlValidationException(ErrorCode.OtherXml, ErrorCode.Error, "Can't To element validation.\n" + LogUtil.stackTraceToString(e), "To validation");
		}
	}
	
	
	/**
	 * Version     - v1.0
	 * Create time - May 7, 2007
	 * Update time - May 7, 2007
	 *
	 * Author      - Sirius
	 *
	 * @param ebxmlMessage
	 * @throws EbmsEbxmlValidationException
	 * 
	 * Error condition
	 * 1. If 'Service''s 'type' attribute is not present, 'Service' element value MUST be a URI.
	 * 2. The Service element of ebXML message does not exist to the CPA.
	 */
	private void serviceValidation(EbxmlMessage ebxmlMessage) throws EbmsEbxmlValidationException {
		try {
			log.info(" Service element validation.");
			
			Service service = ebxmlMessage.getEbxmlHeader().getMessageHeader().getService();
			
			String type = service.getType();
			String value = service.getValue();
			
			log.info("  Service type: " + type + "\tvalue: " + value);
			
			if( MessageUtil.isSystemLevelActionMessage(ebxmlMessage) ) {
				log.info("  Action is Acknowledgment. Check Service is urn:oasis:names:tc:ebxml-msg:service.");
				
				if( !value.equals("urn:oasis:names:tc:ebxml-msg:service") ) {
					log.fatal("   [ Error ] - Service is not urn:oasis:names:tc:ebxml-msg:service. The service of ebXML message : " + value);
					
					throw new EbmsEbxmlValidationException(
							ErrorCode.ValueNotRecognized,
							ErrorCode.Error,
							"Service value not recognized", "Service validation."
							);
				}
			}
			else {
				if( type == null ) {
					log.info("   Service type is null. Check Service element value is URI.");
					
					if( !isValidUrl(value) ) {
						log.fatal("    [ Error ] - Service element value is not URI.");
						
						throw new EbmsEbxmlValidationException(ErrorCode.ValueNotRecognized, ErrorCode.Error, "Service is not uri.", "Service validation");
					}
				}
				
				log.info("  Search Service to the CPA.");
				
				String cpaId = ebxmlMessage.getCPAId();
				String to = ebxmlMessage.getTo();
				
				Cpa cpa = new CpaManagerFactory().createInstance().getCpa(cpaId);
				
				PartyInfo partyInfo = cpa.getPartyInfo(to);
				
				String serviceStr = partyInfo.getService();
				
				if( value.equals(serviceStr) ) 
					log.info("   The service name is same.");
				else {
					log.fatal("   The service name is not same.");
					log.fatal("    Message service: " + value + "\tCPA service: " + serviceStr);
					
					throw new EbmsEbxmlValidationException(
							ErrorCode.ValueNotRecognized  ,
							ErrorCode.Error,
							"Service value not recognized.", "Service validation.");
				}
			}
		}
		catch(SOAPException e) {
			throw new EbmsEbxmlValidationException(ErrorCode.OtherXml, ErrorCode.Error, "Can't Service element validation.\n" + LogUtil.stackTraceToString(e), "Service validation");
		}
	}
	
	/**
	 * Version     - v1.0
	 * Create time - May 9, 2007
	 * Update time - May 9, 2007
	 *
	 * Author      - Sirius
	 *
	 * @param ebxmlMessage
	 * @throws EbmsEbxmlValidationException
	 * 
	 * Error condition
	 * 1. The case that the Action element of ebxml message is not see clearly in the CPA.
	 */
	private void actionValidation(EbxmlMessage ebxmlMessage) throws EbmsEbxmlValidationException {
		log.info(" Action validation.");
		
		if( MessageUtil.isSystemLevelActionMessage(ebxmlMessage) )
			return ;
		
		log.info("  Search Action from CPA.");
		boolean findStatus = false;
		
		String cpaId = ebxmlMessage.getCPAId();
		String ebXMLMessageAction = ebxmlMessage.getAction();
		
		Cpa cpa = new CpaManagerFactory().createInstance().getCpa(cpaId);
		
		String to = ebxmlMessage.getTo();
		
		PartyInfo partyInfo = cpa.getPartyInfo(to);
		
		CanReceive[] canReceiveArray = partyInfo.getCanReceives();
		
		for(int i = 0; i < canReceiveArray.length; i++) {
			CanReceive canReceive = canReceiveArray[i];
			
			String cpaAction = canReceive.getAction();
			
			if( ebXMLMessageAction.equals(cpaAction) ) {
				log.info("   " + ebXMLMessageAction + " == " + cpaAction);
				
				findStatus = true;
				
				break;
			}
			
			log.info("   " + ebXMLMessageAction + " != " + cpaAction);
		}
		
		if( !findStatus ) {
			log.fatal("    [ Error ] - Can't find ebXML message action.");
			throw new EbmsEbxmlValidationException(
					ErrorCode.ValueNotRecognized,
					ErrorCode.Error,
					"No such kinds of Action.",
					"Action validation"); 
		}
	}

	
	/**
	 * Version     - v1.0
	 * Create time - May 9, 2007
	 * Update time - May 9, 2007
	 *
	 * Author      - Sirius
	 *
	 * @param ebxmlMessage
	 * @throws EbmsEbxmlValidationException
	 * 
	 * Error condition
	 * 1. If ebxml message has 'DuplicateElimination' element, check CPA 'DuplicateElimination' attribute is always. 
	 */
	private void duplicateEliminationValidation(EbxmlMessage ebxmlMessage) throws EbmsEbxmlValidationException {
		try {
			log.info(" DuplicateElimination validation.");
			
			boolean messageDuplicateEliminationMode = false;
			boolean cpaDuplicateEliminationMode = false;
			
			if( ebxmlMessage.getEbxmlHeader().getMessageHeader().hasDuplicateElimination() ) 
				messageDuplicateEliminationMode = true;
			
			Cpa cpa = new CpaManagerFactory().createInstance().getCpa(ebxmlMessage.getCPAId());
			
			String from = ebxmlMessage.getFrom();
			
			String action = ebxmlMessage.getAction();
						
			DeliveryChannel deliveryChannel = null;
			
			if( MessageUtil.isSystemLevelActionMessage(ebxmlMessage) ) {
				PartyInfo partyInfo = cpa.getPartyInfo(from);
				
				String defaultMshChannelId = partyInfo.getDefaultMshChannelId();
				
				deliveryChannel = partyInfo.getDeliveryChannel(defaultMshChannelId);
			}
			else {
				deliveryChannel = cpa.getPartyInfo(from).getDeliveryChannelByAction(action);
			}
			
			String duplicateElimination = deliveryChannel.getDuplicateElimination();
			
			if( duplicateElimination.equals("always") )
				cpaDuplicateEliminationMode = true;
			else if( duplicateElimination.equals("never") )
				cpaDuplicateEliminationMode = false;
			else if( duplicateElimination.equals("perMessage") )
				return ;
			
			log.info("  DuplicateElimination Mode - Message: " + messageDuplicateEliminationMode + "\tCPA: " + cpaDuplicateEliminationMode);
			
			if( messageDuplicateEliminationMode != cpaDuplicateEliminationMode ) {
				throw new EbmsEbxmlValidationException(ErrorCode.inconsistent,
						ErrorCode.Error,
						"DuplicateElimination inconsistent with CPA.", "DuplicateElimination validation");
			}
		}
		catch(SOAPException e) {
			throw new EbmsEbxmlValidationException(ErrorCode.OtherXml, ErrorCode.Error, "Can't DuplicateElimination validation.\n" + LogUtil.stackTraceToString(e), "DuplicateElimination validation");
		}
	}

	
	/**
	 * Version     - v1.0
	 * Create time - May 9, 2007
	 * Update time - May 9, 2007
	 *
	 * Author      - Sirius
	 *
	 * @param ebxmlMessage
	 * @throws EbmsEbxmlValidationException
	 * 
	 * Error condition
	 * 1. The ebXML message includes 'SyncReply' element.
	 * 2. The SequenceNumber of ebXML message is smaller the lowest or bigger than top.
	 * 
	 * //TODO not implements..
	 * 3. Limit for saved out-of-sequence messages is reached. 
	 */
	private void messageOrderValidation(EbxmlMessage ebxmlMessage) throws EbmsEbxmlValidationException {
		try {
			log.info(" MessageOrder validation.");
			
			EbxmlHeader ebxmlHeader = ebxmlMessage.getEbxmlHeader();
			
			if( !ebxmlHeader.hasMessageOrder() ) {
				log.info("  MessageOrder does not exist.");
				
				return ;
			}
			
			MessageOrder messageOrder = ebxmlHeader.getMessageOrder();
			SequenceNumber sequenceNumber = messageOrder.getSequenceNumber();
			int sequenceNumberInt = Integer.parseInt(sequenceNumber.getValue());
			
			if( EbxmlConstants.MIN_MESSAGE_ORDER > sequenceNumberInt || EbxmlConstants.MAX_MESSAGE_ORDER < sequenceNumberInt ) {
				log.fatal("  [ Error ] - Current sequence number: " + sequenceNumberInt + "\tSequence number is smaller than " + EbxmlConstants.MIN_MESSAGE_ORDER + " or bigger than " + EbxmlConstants.MAX_MESSAGE_ORDER + ".");
				throw new EbmsEbxmlValidationException(
						ErrorCode.inconsistent,
						ErrorCode.Error,
						"Sequence number is smaller than " + EbxmlConstants.MIN_MESSAGE_ORDER + " or bigger than " + EbxmlConstants.MAX_MESSAGE_ORDER + ".", "MessageOrder validation");
			}
			
			if( ebxmlMessage.getEbxmlHeader().hasSyncReply() ) {
				throw new EbmsEbxmlValidationException(
						ErrorCode.inconsistent,
						ErrorCode.Error,
						"Sync and MessageOrder element can't appear same time that message.", "MessageOrder validation");
			}
		}
		catch(SOAPException e) {
			throw new EbmsEbxmlValidationException(ErrorCode.OtherXml, ErrorCode.Error, "Can't DuplicateElimination validation.\n" + LogUtil.stackTraceToString(e), "DuplicateElimination validation");
		}
	}

	
	/**
	 * Version     - v1.0
	 * Create time - May 9, 2007
	 * Update time - May 9, 2007
	 *
	 * Author      - Sirius
	 *
	 * @param ebxmlMessage
	 * @throws EbmsEbxmlValidationException
	 * 
	 * Error condition
	 * 1. If ebXML message has AckRequest element, check ackRequested attribute of CPA is inconsist.
	 * 2. If ebXML message has AckRequest element, check AckSignatureRequested attribute of AckRequest of ebXML message is inconsist.
	 */
	private void ackRequestedValidation(EbxmlMessage ebxmlMessage) throws EbmsEbxmlValidationException {
		try {
			log.info(" AckRequested validation.");			
			
			boolean messageAckRequested = false;
			
			if( ebxmlMessage.getEbxmlHeader().hasAckRequested() ) 
				messageAckRequested = true;
			
			String cpaId = ebxmlMessage.getCPAId();
			String from = ebxmlMessage.getFrom();
			String action = ebxmlMessage.getAction();
			
			// Acknowledgment, ErrorMessage AckRequested  ü ⿡  з ó.
			if( messageAckRequested ) {
				if( action.equals("Acknowledgment") ) {
					log.fatal("  [ Error ] - Acknowledgment ebXML message can't contain AckRequested element.");
					
					throw new EbmsEbxmlValidationException(ErrorCode.inconsistent,
							ErrorCode.Error,
							"Acknowledgment ebXML message can't contain AckRequested element.", "AckRequested validation");
				}
				else if( action.equals("ErrorMessage") ) {
					log.fatal("  [ Error ] - ErrorMessage ebXML message can't contain AckRequested element.");
					
					throw new EbmsEbxmlValidationException(ErrorCode.inconsistent,
							ErrorCode.Error,
							"ErrorMessage ebXML message can't contain AckRequested element.", "AckRequested validation");
				}
			}
			// Acknowledgment, ErrorMessage AckRequested   ʾұ   մϴ.
			else {
				if( action.equals("Acknowledgment") || action.equals("ErrorMessage") ) 
					return ;
			}
			
			Cpa cpa = new CpaManagerFactory().createInstance().getCpa(cpaId);
			
			PartyInfo partyInfo = cpa.getPartyInfo(from);
			
			DeliveryChannel deliveryChannel = null;
			
			if( MessageUtil.isSystemLevelActionMessage(ebxmlMessage) ) {
				String defaultMshChannelId = partyInfo.getDefaultMshChannelId();
				
				deliveryChannel = partyInfo.getDeliveryChannel(defaultMshChannelId);
			}
			else {
				deliveryChannel = cpa.getPartyInfo(from).getDeliveryChannelByAction(action);
			}
			
			String cpaAckRequested = deliveryChannel.getAckRequested();
			
			log.info("  AckRequested Mode - Message: " + messageAckRequested + "\tCPA: " + cpaAckRequested);
			
			if( messageAckRequested ) {
				if( cpaAckRequested.equals("never") ) {
					throw new EbmsEbxmlValidationException(ErrorCode.inconsistent,
							ErrorCode.Error,
							"AckRequested inconsistent with CPA.", "AckRequested validation");
				}
			}
			else {
				if( cpaAckRequested.equals("always") ) {
					throw new EbmsEbxmlValidationException(ErrorCode.inconsistent,
							ErrorCode.Error,
							"AckRequested inconsistent with CPA.", "AckRequested validation");
				}
			}
			
			if( messageAckRequested ) {
				log.info("  Check AckSignatureRequested.");
				boolean messageAckSignatureRequested = false;
				
				messageAckSignatureRequested = ebxmlMessage.getEbxmlHeader().getAckRequested().getSigned();
				
				String cpaAckSignatureRequested = deliveryChannel.getAckSignatureRequested();
				
				log.info("  AckSignatureRequested Mode - Message: " + messageAckSignatureRequested + "\tCPA: " + cpaAckSignatureRequested);
				
				if( messageAckSignatureRequested ) {
					if( cpaAckSignatureRequested.equals("never") || cpaAckSignatureRequested.equals("perMessage") ) {
						throw new EbmsEbxmlValidationException(ErrorCode.inconsistent,
								ErrorCode.Error,
								"AckSignatureRequested inconsistent with CPA.", "AckRequested validation");
					}
				}
				else {
					if( cpaAckSignatureRequested.equals("always") ) {
						throw new EbmsEbxmlValidationException(ErrorCode.inconsistent,
								ErrorCode.Error,
								"AckSignatureRequested inconsistent with CPA.", "AckRequested validation");
					}
				}
			}
		}
		catch(SOAPException e) {
			throw new EbmsEbxmlValidationException(ErrorCode.OtherXml, ErrorCode.Error, "Can't AckRequested validation.\n" + LogUtil.stackTraceToString(e), "AckRequested validation");
		}
	}

	
	/**
	 * Version     - v1.0
	 * Create time - May 9, 2007
	 * Update time - May 9, 2007
	 *
	 * Author      - Sirius
	 *
	 * @param ebxmlMessage
	 * @throws EbmsEbxmlValidationException
	 * 
	 * Error condition
	 * 1. The case that the TimeToLive of ebXML message, after current time. 
	 */
	private void timeToLiveValidation(EbxmlMessage ebxmlMessage) throws EbmsEbxmlValidationException {
		log.info(" TimeToLive validation.");
		
		String timeToLive = ebxmlMessage.getTimeToLive();
		
		if( timeToLive == null ) {
			log.info("  TimeToLive does not exist.");
			
			return ;
		}
		
		Date timestamp = MessageUtil.utcToDate(timeToLive);
		
		if( timestamp.before(new Date()) ) {
			log.fatal("   Message TimeToLive: " + timestamp + "\tCurrent time: " + new Date());
			
			throw new EbmsEbxmlValidationException(
					ErrorCode.TimeToLiveExpired,
					ErrorCode.Error,
					"Invalid TimeToLive.", "TimeToLive validation");
		}
	}

	

	
	/**
	 * Version     - v1.0
	 * Create time - May 7, 2007
	 * Update time - May 7, 2007
	 *
	 * Author      - Sirius
	 *
	 * @param ebxmlMessage
	 * @throws EbmsEbxmlValidationException
	 * 
	 * Error condition
	 * 1. If ebxml message has 'SyncReply' element, check CPA 'SyncReplyMode' attribute is always.
	 */
	private void syncValidation(EbxmlMessage ebxmlMessage) throws EbmsEbxmlValidationException {
		try {
			log.info(" SyncReply validation.");
			
			boolean messageSyncReplyMode = false;
			boolean cpaSyncReplyMode = false;
			
			EbxmlHeader ebxmlHeader = (EbxmlHeader)ebxmlMessage.getEbxmlHeader();
			
			if( ebxmlHeader.hasSyncReply() )
				messageSyncReplyMode = true;
			
			Cpa cpa = new CpaManagerFactory().createInstance().getCpa(ebxmlMessage.getCPAId());
			String syncReplyMode = null;
			
			if( MessageUtil.isSystemLevelActionMessage(ebxmlMessage) ) {
				PartyInfo partyInfo = cpa.getPartyInfo(ebxmlMessage.getFrom());
				
				String defaultMshChannelId = partyInfo.getDefaultMshChannelId();
				
				syncReplyMode = partyInfo.getDeliveryChannel(defaultMshChannelId).getSyncReplyMode();
			}
			else {
				syncReplyMode = cpa.getPartyInfo(ebxmlMessage.getFrom()).getDeliveryChannelByAction(ebxmlMessage.getAction()).getSyncReplyMode();
			}
			
			if( syncReplyMode.equals("none") ) 
				cpaSyncReplyMode = false;
			else if( syncReplyMode.equals("mshSignalsOnly") || syncReplyMode.equals("signalsOnly") || syncReplyMode.equals("responseOnly") || syncReplyMode.equals("signalsAndResponse") )
				cpaSyncReplyMode = true;
			
			log.info("  SyncReply Mode - Message: " + messageSyncReplyMode + "\tCPA: " + syncReplyMode);
			
			if( messageSyncReplyMode != cpaSyncReplyMode ) {
				throw new EbmsEbxmlValidationException(ErrorCode.inconsistent,
						ErrorCode.Error,
						"SyncReplyMode inconsistent with CPA.", "SyncReplyMode validation");
			}
		}
		catch(SOAPException e) {
			throw new EbmsEbxmlValidationException(ErrorCode.OtherXml, ErrorCode.Error, "Can't SyncReply validation.\n" + LogUtil
					.stackTraceToString(e), "SyncReply validation");
		}
	}
	
	/**
	 * Version     - v1.0
	 * Create time - May 8, 2007
	 * Update time - May 8, 2007
	 *
	 * Author      - Sirius
	 *
	 * @param url
	 * @return
	 * 
	 * Validate the url.
	 */
	private boolean isValidUrl(String url) {
		try {
			new URL(url);
			
			return true;
		}
		catch (MalformedURLException e) {
			return false;
		}
	}
	
}
