/*
 * Copyright (c) Sybase, Inc. 2002   
 * All rights reserved.                                    
 * 
 * This file contains Original Code and/or Modifications of 
 * Original Code as defined in and that are subject to the  
 * Sybase Public Source License version 1.0 (the           
 * "License").  You may not use this file except in         
 * compliance with the License.  Please obtain a copy of    
 * the License at http://www.sybase.com/developer/opensource 
 * and read it before using this file.                               
 * 
 * The Original Code and all software distributed under the 
 * License are distributed on an 'AS IS' BASIS, WITHOUT     
 * WARRANTY OF ANY KIND, EITHER EXPRESS OR IMPLIED, AND     
 * SYBASE HEREBY DISCLAIMS ALL SUCH WARRANTIES, INCLUDING  
 * WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,    
 * FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR     
 * NON-INFRINGEMENT.                                        
 * 
 * Please see the License for the specific language.         
 */
package org.kldp.ebmsh.message.util;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;

import javax.activation.DataSource;
import javax.activation.FileTypeMap;
import javax.mail.internet.ContentType;
import javax.mail.internet.MimeUtility;
import javax.mail.internet.ParseException;

/**
 * This class implements a typed DataSource from
 * 	an InputStream,
 *	a byte array,
 * 	a String
 *  a File.
 *
 * @author Cliff Collins
 */
public class ByteArrayDataSource
    implements DataSource, java.io.Serializable
{
    //private static com.sybase.bis.logging.Logger _log =
    // com.sybase.bis.logging.Logger.getTrpLogger(ByteArrayDataSource.class);
    
	// The data will be either store in a byte[] (if small) or
	// as a file on disk if the constructor used was File based, or the
	// input stream is very large.
    private byte[]		_data;			// data source
    private	File		_file;			// data source
    private String		_type;			// content-type
    
    // this is used for header formating. 
    private byte		_fileFirstByte;
    private boolean		_bCreatedTmpFile;	// this message created a tmp file...should we clean it up?
    
    // Save the last read of this file so we can make sure 
    // it is cleared and be able to remove the file.
    //
    // NOTE: JavaMail MIME opens data source via the getInputStream() but never
    // closes it which prevents cleanup of the file.
	// NOTE: we don't save the last OutputStream because it was the callers
    // responsibility to close this. 
    private	InputStream	_lastFileInputStream;
    
    // these fields are all extra fields to help stats and understand the nature of 
    // the bytes. These are set when it is useful but they are not required in 
    // any way.
    private	String		_soapAction;	// soap action header
    private	String		_encoding;		// if this is encoded this might be base64.
    private	boolean		_bEncryptedData;// true if the data is encrypted. It is up to the creator to set this.
    private	boolean		_bSignedData;	// true if the data is signed. It is up to the creator to set this.
    private	boolean		_bSSLData;		// true if the data is was or will be sent over SSL. It is up to the creator to set this.
    private	boolean		_bEncapsulatedMsg;// true if the data is encapsulating another message. It is up to the creator to set this.

    /**
     * Create a datasource but with no data. The intension is to use the getOutputStream()
     * to fill the data into this source.
     *
     * @param type Content-Type
     */
    public ByteArrayDataSource(String type)
    {
        _type = type;  
    }
    
    /**
     * Create a datasource from a File. If the Content-Type parameter is null,
     * the type will be derived from the filename extension.
     *
     * @param ff File object
     * @param type Content-Type
     * @param soapAction the Soap Action header
     */
    public ByteArrayDataSource(File ff, String type)
        throws IOException
    {
    	if ( (ff == null) || (ff.exists() == false) || (ff.isFile() == false) )
    	{
    		throw new IOException("File Not Found: " + ff);
    	}
        
    	_file = ff;
        _type = type;  

        if (_type == null)
        {
            _type = FileTypeMap.getDefaultFileTypeMap().getContentType(ff);
        }
    }

    /**
     * Create a datasource from an input stream.
     *
     * @param is InputStream
     * @param type Content-Type
     * @param soapAction the Soap Action header
     * @param streamToTmpDiskSize -1 for never stream to disk otherwise set this to the number
     * of bytes to keep in memory verse writing to a temp file on disk for reference. A good example
     * of this might be 50000 (50k).
     */
    public ByteArrayDataSource(InputStream is, String type, int streamToTmpDiskSize)
        throws IOException
    {
        _type = type;
        
        ByteArrayOutputStream	bos = new ByteArrayOutputStream();
        FileOutputStream		out = null;
		try
		{
    		boolean					bStreamToDisk = (streamToTmpDiskSize > 0);
        	int						count = streamToTmpDiskSize;

        	byte	buf[] = new byte[4096];
        	int		len;
        	while (true)
        	{
            	len = is.read(buf);
            	if (len < 0)
            	{
                	break;
            	}
				
				if (out == null)
				{
					count -= len;
            		bos.write(buf, 0 , len);
	            	
            		if ( bStreamToDisk && (count < 0) )
            		{
            			// save to disk instead.
            			_bCreatedTmpFile = true;
                    	_file = File.createTempFile("ebxml-", null, null);
                        // _log.info("trp.ebxml.log.0001", "new InputStream file=" + _file, new Exception("New file from here"));
                    	out = new FileOutputStream(_file);
                    	byte[] tmp = bos.toByteArray();
                    	_fileFirstByte = tmp[0];
                    	out.write(tmp);
              		}
              	}
              	else
              	{
            		out.write(buf, 0 , len);
              	}
        	}

			if (out != null)
			{
                out.flush();
                out.close();
                out = null;
                // NOTE: _file should be set.
            }
			else
			{
        		_data = bos.toByteArray();
        	}
		}
        finally
        {
			if (out != null)
			{
                out.close();
            }
    		bos.close();
    	}
    }

    /**
     * Create a datasource from a byte array.
     *
     * @param data byte array
     * @param type Content-Type
     */
    public ByteArrayDataSource(byte[] data, String type)
    {
        _type = type;
        _data = data;
    }

    /**
     * Create a datasource from a String. This method defaults to
     * a String encoding of iso-8859-1. For a different encoding,
     * specify a Mime "charset" in the Content-Type parameter.
     *
     * @param data byte array
     * @param type Content-Type
     */
    public ByteArrayDataSource(String data, String type)
    {
        try
        {
            _type = type;
            _data = data.getBytes(getEncoding(_type,
                SOAPConstants.HEADERVAL_DEFAULT_CHARSET));
        }
        catch (UnsupportedEncodingException uex)
        {
        }
    }
    
    /**
     * Clean up any tmp files we created and close any references 
     * to this object.
     */
    public void destroy()
    {
    	if (_file != null)
    	{
            // close any possible open connections to this file
            // so we can delete it.
            if (_lastFileInputStream != null)
            {
                try
                {
                    _lastFileInputStream.close();
                }
                catch (Exception ee)
                {
                }
            }
    		if (_bCreatedTmpFile)
    		{
                // _log.info("trp.ebxml.log.0001", "delete Ofile=" + _file, new Exception("Delete from here"));
    			_file.delete();
    		}
    	}
    }
    
    /**
     * Read a InputStream into a byte array in memory
     */
	private byte[] readInputStreamIntoBytes(InputStream is, boolean bCloseStream)
        throws IOException
	{
		ByteArrayOutputStream bos = new ByteArrayOutputStream();
		try
		{
        	byte	buf[] = new byte[4096];
        	int		len;
        	while (true)
        	{
            	len = is.read(buf);
            	if (len < 0)
            	{
                	break;
            	}

            	bos.write(buf, 0 , len);
        	}

        	return bos.toByteArray();
		}
        finally
        {
        	if (bCloseStream)
        	{
        		is.close();
    		}
    		bos.close();
    	}
	}

    /**
     * Return an InputStream to read the content.
     *
     * @return an InputStream with the content
     */
    public InputStream getInputStream()
        throws IOException
    {
        if (_data != null)
        {
        	return new ByteArrayInputStream(_data);
        }
        else if (_file != null)
        {
        	_lastFileInputStream = new FileInputStream(_file);
        	return _lastFileInputStream;
        }
        else
        {
            throw new IOException("No data.");
        }
    }

    /**
     * This DataSource cannot return an OutputStream, so this method is not
     * implemented.
     */
    public OutputStream getOutputStream()
        throws IOException
    {
    	return this.getOutputStream((_file == null));
    }
    
    /**
     * This DataSource cannot return an OutputStream, so this method is not
     * implemented.
     */
    public OutputStream getOutputStream(boolean bStoreAsBytesData)
        throws IOException
    {
    	OutputStream out = null;
    	if (bStoreAsBytesData)
    	{
            out = new ByteArrayOutputStream();
    		if (_data != null)
    		{
        		out.write(_data);	// write the old data first for appending.
        	}
		}
		else
		{
			if (_file == null)
			{
            	_bCreatedTmpFile = true;
                _file = File.createTempFile("ebxml-", null, null);
                // _log.info("trp.ebxml.log.0001", "new OutputStream file=" + _file, new Exception("New file from here"));
            	out = new FileOutputStream(_file);
            }
            else
            {
            	out = new FileOutputStream(_file.getCanonicalPath(), true);
            }
		}
		
		return out;
    }
    
    /**
     * Since we stream to a byte array we have to know when it is done so we can
     * get the byte[]. This is kinda kludgy but is the only way to support this call with
     * for a byte array without writing to a file or something. Note: This method 
     * will close and flush the stream.
     *
     * @param out The OutputStream returned from getOutputStream.
     */
    public void doneWithOutputStream(OutputStream out)
        throws IOException
    {
    	if (out != null)
    	{
    		out.flush();
    		out.close();
    		if (out instanceof ByteArrayOutputStream)
    		{
    			_data = ((ByteArrayOutputStream)out).toByteArray();
    		}
    	}
    }

    /**
     * Get the content type.
     *
     * @return Content-Type string
     */
    public String getContentType()
    {
        return _type;
    }

    /**
     * Set the content type.
     *
     * @param Content-Type string
     */
    public void setContentType(String type)
    {
        _type = type;
    }

    /**
     * Get the Soap action header
     *
     * @return soapAction
     */
    public String getSoapAction()
    {
        return _soapAction;
    }
    /**
     * Set the Soap action header
     *
     * @param soapAction
     */
    public void setSoapAction(String soapAction)
    {
        _soapAction = soapAction;
    }

    /**
     * Get the Content-Transfer-Encoding header
     *
     * @return Content-Transfer-Encoding (ex "base64")
     */
    public String getEncoding()
    {
        return _encoding;
    }
    /**
     * Set the Content-Transfer-Encoding header
     *
     * @param Content-Transfer-Encoding (base64)
     */
    public void setEncoding(String encoding)
    {
        _encoding = encoding;
    }

    /**
     * @return boolean of the data bytes are encrypted.
     */
    public boolean isEncryptedData()
    {
        return _bEncryptedData;
    }
    /**
     * After reading the stream we often can determine if the bytes are encrypted.
     *
     * @param bEncryptedData is this data encrypted.
     */
    public void setEncryptedData(boolean bEncryptedData)
    {
        _bEncryptedData = bEncryptedData;
    }

    /**
     * @return boolean of the data bytes are Encapsulating another message.
     */
    public boolean isEncapsulatedMsg()
    {
        return _bEncapsulatedMsg;
    }
    /**
     * After reading the stream we often can determine if the bytes are Encapsulating another message.
     *
     * @param bEncapsulatedMsg is this data Encapsulating another message.
     */
    public void setEncapsulatedMsg(boolean bEncapsulatedMsg)
    {
        _bEncapsulatedMsg = bEncapsulatedMsg;
    }

    /**
     * @return boolean of the data bytes are signed.
     */
    public boolean isSignedData()
    {
        return _bSignedData;
    }
    /**
     * After reading the stream we often can determine if the bytes contain a signature.
     *
     * @param bSignedData is this data signed.
     */
    public void setSignedData(boolean bSignedData)
    {
        _bSignedData = bSignedData;
    }

    /**
     * @return boolean of the data bytes are to sent/received over SSL.
     */
    public boolean isSSLData()
    {
        return _bSSLData;
    }
    /**
     * After reading the stream we can determine if the bytes was sent over ssl.
     *
     * @param bSignedData is this data signed.
     */
    public void setSSLData(boolean bSSL)
    {
        _bSSLData = bSSL;
    }

    /**
     * getName()
     * NOTE: is not implemented for byte streams.
     */
    public String getName()
    {
    	if (_file != null)
    	{
    		return _file.getName();
    	}
        return "";	// no name for a byte array.
    }

    /**
     * Write the content to an OutputStream.
     *
     * @param os OutputStream to write the entire content to
     */
    public void writeTo(OutputStream os)
        throws IOException
    {
    	if (_data != null)
    	{
        	os.write(_data);
        }
        else if (_file != null)
        {
    		InputStream 	is = new FileInputStream(_file);
    		try
    		{
        		byte		buf[] = new byte[4096];
        		int			len;
	        	
        		while (true)
        		{
            		len = is.read(buf);
            		if (len < 0)
            		{
                		break;
            		}

            		os.write(buf, 0 , len);
        		}
			}
        	finally
        	{
        		is.close();
        	}
        }
    }

    /**
     * Write the content type string to an OutputStream.
     *
     * @param os OutputStream to write the content type to
     */
    public void headerWriteTo(OutputStream os)
        throws IOException
    {
        if (_type != null)
        {
            os.write(SOAPConstants.HEADER_CONTENT_TYPE.getBytes());
            os.write(':');
            os.write(' ');
            os.write(_type.getBytes());
            os.write(SOAPConstants.CR.getBytes());
        }
        if (_encoding != null)
        {
            os.write(SOAPConstants.HEADER_CONTENT_ENCODING.getBytes());
            os.write(':');
            os.write(' ');
            os.write(_encoding.getBytes());
            os.write(SOAPConstants.CR.getBytes());
        }
        if (_soapAction != null)
        {
            os.write(SOAPConstants.HEADER_SOAP_ACTION.getBytes());
            os.write(':');
            os.write(' ');
            os.write(_soapAction.getBytes());
            os.write(SOAPConstants.CR.getBytes());
        }
    }

    /**
     * Return the content as a byte array.
     *
     * @return byte array with the content
     */
    public byte[] toByteArray()
        throws IOException
    {
    	if ( (_data == null) && (_file != null) )
    	{
    		// read the bytes from the file into a byte array
    		InputStream is = new FileInputStream(_file);
    		_data = readInputStreamIntoBytes(is, true);
    	}
        return _data;
    }
    
    /**
     * Return the content as a byte array and remove and starting and trailing 
     * white space.
     *
     * @return byte array with the content
     */
    public byte[] toByteArrayTrim()
        throws IOException
    {
        try
        {
            String content = getText();

            if (content != null)
            {
                return content.trim().getBytes(getEncoding(_type,
                    SOAPConstants.HEADERVAL_DEFAULT_CHARSET));
            }
        }
        catch (UnsupportedEncodingException uex)
        {
        }

        return _data;
    }

    /** 
     * Return true if the first character in the byte stream is a new line.
     */
    public boolean isFirstByteLF()
    {
    	if ( (_data != null) && (_data.length > 0) )
    	{
    		return ((_data[0] == '\n') || (_data[0] == '\r'));
    	}
    	else
    	{
    		return ((_fileFirstByte == '\n') || (_fileFirstByte == '\r'));
    	}
    }

    /**
     * Return the number of bytes in the content.
     *
     * @return size of the byte array, or -1 if not set.
     */
    public int getSize()
    {
        if (_data != null)
        {
            return _data.length;
        }
        else if (_file != null)
        {
        	return (int)_file.length();
        }
        else
        {
            return -1;
        }
    }

    /**
     * Return the content as a String. The Content-Type "charset" parameter
     * will be used to determine the encoding, and if that's not available or
     * invalid, "iso-8859-1".
     *
     * @return a String with the content
     */
    public String getText()
        throws IOException
    {
    	if (_data == null)
    	{
    		toByteArray();
    	}
    	if (_data == null)
    	{
    		return null;
    	}
    	
        try
        {
            return new String(_data, getEncoding(_type,
                SOAPConstants.HEADERVAL_DEFAULT_CHARSET));
        }
        catch (UnsupportedEncodingException uex)
        {
            try
            {
                return new String(_data, SOAPConstants.HEADERVAL_DEFAULT_CHARSET);
            }
            catch (UnsupportedEncodingException uex1)
            {
                return null;
            }
        }
    }

    /**
     * Extract a Java encoding from a Content-Type's MIME character set.
     */
    public static String getEncoding(String type, String defaultEncoding)
    {
        String encoding = null;

        try
        {
            if (type != null && !type.equals(""))
            {
                encoding = new ContentType(type).getParameter("charset");
            }
        }
        catch(ParseException pe)
        {
        }

        if (encoding == null)
        {
            encoding = defaultEncoding;
        }
        else
        {
            encoding = MimeUtility.javaCharset(encoding);
        }
        return encoding;
    }
}
